import datetime
import os
from pathlib import Path
from robot.api import logger
from time import sleep
from .testrail import APIClient

TEST_TEMPLATE_ID = 1
TEST_TYPE_ID = 3
PRIORITY_DO_NOT_TEST = 1
PRIORITY_CRITICAL = 4

TEST_STATUS_PASSED = 1
TEST_STATUS_FAILED = 5
TEST_STATUS_BLOCKED = 2
TEST_STATUS_UNTESTED = 3
TEST_STATUS_RETEST = 4
TEST_STATUS_CUSTOM_6 = 6
TEST_STATUS_CUSTOM_7 = 7

VERSION = "1.0.0"

class TestrailLibrary:
    
    ROBOT_LIBRARY_VERSION = VERSION
    ROBOT_LIBRARY_SCOPE = "GLOBAL"
    
    class TestCaseResult:

        def __init__(self):
            self.include_all = True
            self.description = None
            self.project_id = None
            self.run_id = None
            self.run_name = None
            self.section_id = None
            self.section_name = None
            self.suite_id = None
            self.suite_name = None
            self.test_id = None
            self.test_name = None
            self.elapsed = None
            self.comment = None
            self.status_id = TEST_STATUS_UNTESTED
            self.template_id = None
            self.type_id = None
            self.priority_id = None

        def __setitem__(self, item, value):
            setattr(self, item, value)

        def __getitem__(self, key):
            return getattr(self, key)

    def __init__(self, server=None, username=None, key=None):

        self.server = server
        self.client = None
        self.expect_duplicates = False
        if server:
            self.client = APIClient(server)
            self.client.user = username
            self.client.password = key
        self.test_case_result = self.TestCaseResult()

    def __handle_duplicates(self, msg):
        if self.expect_duplicates:
            logger.warn(msg)
        else:
            raise DuplicateException(msg)

    def connect_to_testrail(self, server, username, key):
        self.client = APIClient(server)
        self.client.user = username
        self.client.password = key
        self.__check_credentials()
    
    def set_expect_duplicates(self, state):
        self.expect_duplicates = state

    def set_test_case_result(self, **kwargs):
        for key, value in kwargs.items():
            self.test_case_result[f"{key}"] = value

    def __check_credentials(self):
        if not self.client.user or not self.client.password:
           raise AssertionError('Both testrail username and api key should be passed to the library. \
           Please use the "set testrail credentials" keyword to assign them.')

    def __get_with_offset(self, base_url, attribute, **kwargs):
        
        offset = kwargs.get('offset') or 0
        try:
            offset = int(offset)
        except (ValueError):
            offset = 0
        limit = kwargs.get('limit') or 250
        try:
            limit = int(limit)
        except (ValueError):
            limit = 250
        results = []
        url = base_url
        for key, value in kwargs.items():
            if key != 'offset':
                url = f'{url}&{key}={value}'
        count = 0
        while True and count < 50:
            response = self.client.send_get(f'{url}&offset={offset}')
            if not response or not response.get(attribute):
                break
            assert response.get('offset') == offset, f"{response.get('offset')} != {offset}"
            if limit:
                assert response.get('limit') == limit, f"{response.get('limit')} != {limit}"
            results.extend(response.get(attribute))
            offset += limit
            if response.get('size') and response.get('size') < 250 or limit < 250:
                break
            count += 1
            sleep(0.5)
        if len(results) > 50:
            info = f'{results[:20]}\n...truncated...\n{results[:-20]}'
        else:
            info = results
        logger.info(info)  
        return results 

    # PROJECTS
    def get_project_by_name(self, project_name):
        projects = self.get_projects()
        if not isinstance(projects, list): 
            projects = []   
        project = list(filter(lambda x:x.get('name')==project_name, projects))
        logger.info(project)
        if project and len(project) == 1:
            return project[0]
        elif project and len(project) > 1:
            #logger.info(project)
            raise Exception(f'More than 1 projects found with name {project_name}')
        else:
            pass
        return {}

    def get_projects(self, **kwargs):
        url = f'get_projects/'
        return self.__get_with_offset(url, "projects", **kwargs)

    def get_project(self, project_id):
        response = self.client.send_get(f'get_project/{project_id}')
        logger.info(response)
        return response
    
    def add_project(self, project_name, description, show=True, suite_mode=3):
        data = {
            "name": project_name,
            "announcement": description,
            "show_announcement": show,
            "suite_mode": suite_mode
        }
        url = 'add_project'
        response = self.client.send_post(url, data=data)
        return response

    # def update_project(self, project_id, project_name=None, description=None, show=True, is_completed=False):
    #     data = {
    #         "show_announcement": show,
    #         "is_completed": is_completed
    #     }
    #     if project_name:
    #         data["name"] = project_name
    #     if description:
    #         data['announcement'] = description
    #     url = f'update_project/{project_id}'
    #     response = self.client.send_post(url, data=data)
    #     return response

    # def delete_project(self, project_id):
    #     url = f'delete_project/{project_id}'
    #     self.client.send_post(url, data={})

    # TEST SUITES
    def get_suite(self, suite_id):
        url = f'get_suite/{suite_id}'
        response = self.client.send_get(url)
        return response

    def get_suites_in_project(self, project_id):
        url = f'get_suites/{project_id}'
        response = self.client.send_get(url)
        return response

    def get_suites_by_name(self, project_id, suite_name, expect_duplicates=False):
        suites = self.get_suites_in_project(project_id)
        if not isinstance(suites, list): suites = []
        suite = list(filter(lambda x: x.get('name') == suite_name, suites))
        if suite and len(suite) > 1:
            msg = f'More than 1 suite found with name {suite_name} for project {project_id}'
            self.__handle_duplicates(msg)
        return suite

    def add_suite(self, project_id, suite_name, description):
        suite = self.get_suites_by_name(project_id, suite_name)
        if suite:
            return suite[0]
        else:
            url = f'add_suite/{project_id}'
            # logger.console(url)
            data = {
                "name": suite_name,
                "description": description
            }
            response = self.client.send_post(url, data=data)
            return response
    
    def update_suite(self, suite_id, name=None, description=None):
        url = f'update_suite/{suite_id}'
        data = {}
        if name:
            data['name'] = name
        if description:
            data['description'] = description
        if data:
            response = self.client.send_post(url, data=data)
            return response
        return {}

    # def delete_suite(self, suite_id):
    #     url = f'delete_suite/{suite_id}'
    #     response = self.client.send_post(url, data={})
    #     return response
    
    # TEST SECTIONS
    def get_sections(self, project_id, suite_id, **kwargs):
        url = f'get_sections/{project_id}&suite_id={suite_id}'
        return self.__get_with_offset(url, "sections", **kwargs)

    def get_section(self, section_id):
        url = f'get_section/{section_id}'
        response = self.client.send_get(url)
        return response

    def get_sections_by_name(self, project_id, suite_id, section_name):
        sections = self.get_sections(project_id, suite_id)
        if not isinstance(sections, list): sections = []
        section = list(filter(lambda x: x.get('name') == section_name, sections))
        if section and len(section) > 1:
            msg = f'Found more than 1 section with name {section_name} for ' \
                  f'project {project_id}, suite {suite_id}...returning first section'
            self.__handle_duplicates(msg)
        return section
        
    def add_section(self, project_id, name, description, suite_id, parent_id=None):
        section = self.get_sections_by_name(project_id, suite_id, name) 
        if section:
            return section[0]
        else:
            url = f'add_section/{project_id}'
            data = {
                "name": name,
                "description": description,
                "suite_id": suite_id
            }
            if parent_id:
                data["parent_id"] = parent_id
            response = self.client.send_post(url, data=data)
            return response

    def add_section_to_suite(self, project_id, name, description, suite_name, parent_id=None):
        suite = self.get_suite_by_name(project_id, name)
        if suite:
            response = self.add_section(name=name, description=description, 
                             project_id=project_id, 
                             suite_id=suite.get("id"),
                             parent_id=parent_id
                             )
            return response
        logger.error(f'Failed to add section "{name}". Suite {suite_name} not found in project {project_id}')
        return {}
    
    def update_section(self, section_id, name=None, description=None):
        url = f'update_section/{section_id}'
        data = {}
        if name:
            data['name'] = name
        if description:
            data['description'] = description
        if data:
            response = self.client.send_post(url, data=data)
            return response
        return {}

    # def delete_section(self, section_id):
    #     url = f'delete_section/{section_id}'
    #     response = self.client.send_post(url, data={})
    #     return response

    # CASES
    def get_case(self, test_id, **kwargs):
        url = f'get_case/{test_id}'
        response = self.client.send_get(url, **kwargs)
        logger.info(response)
        return response

    def get_cases(self, project_id, suite_id, **kwargs):
        url = f'get_cases/{project_id}&suite_id={suite_id}'
        return self.__get_with_offset(url, "cases", **kwargs)
    
    def get_cases_by_name(self, project_id, suite_id, section_id, title):
        cases = self.get_cases(project_id, suite_id=suite_id)
        if not isinstance(cases, list): cases = []
        case = list(filter(lambda x: x.get('title') == title and x.get('section_id') == section_id,
                                           cases))
        if case and len(case) > 1:
            msg = f'More than 1 case found with name {title} for project {project_id} ' \
                    f'suite {suite_id} and section {section_id}: {case}'
            self.__handle_duplicates(msg)
        return case

    def add_case(self, project_id, suite_id, title, section_id=None, **kwargs):
        section_name = kwargs.get('section_name') or 'New Section'
        if not section_id:
            sections = self.get_sections_by_name(project_id, suite_id, section_name)
            if not sections:
                section = self.add_section(project_id, name=section_name, description=kwargs.get('description') or 
                'New Section', suite_id=suite_id, parent_id=kwargs.get('parent_id'))
            else:
                if len(sections) > 1:
                    return {}
                section = sections[0]
            section_id = section.get('id')
        case = self.get_cases_by_name(project_id, suite_id, section_id, title) 
        if case: 
            if len(case) > 1:
                return {}
            response = case[0]
        else:
            response = self.add_case_to_section(section_id, title, **kwargs)
        logger.info(response)
        return response

    def add_case_to_section(self, section_id, title, **kwargs):
        template_id = kwargs.get('template_id') or TEST_TEMPLATE_ID
        type_id = kwargs.get('type_id') or TEST_TYPE_ID
        priority_id = kwargs.get('priority_id') or PRIORITY_CRITICAL
        url = f'add_case/{section_id}'
        data = {
            "title": title,
            "template_id": template_id,
            "type_id": type_id,
            "priority_id": priority_id
        }
        response = self.client.send_post(url, data=data)
        logger.info(response)
        return response

    # TESTS
    def get_test(self, test_id):
        url = f'get_test/{test_id}'
        response = self.client.send_get(url)
        logger.info(response)
        return response

    def get_tests(self, run_id, **kwargs):
        url = f'get_tests/{run_id}'
        return self.__get_with_offset(url, "tests", **kwargs)
    
    def get_tests_by_case_name(self, run_id, title):
        tests = self.get_tests(run_id)
        if not isinstance(tests, list): tests = []
        tests_for_title = list(filter(lambda x: x.get('title') == title, tests))
        return tests_for_title

    # RUNS
    def get_runs(self, project_id, **kwargs):
        url = f'get_runs/{project_id}'
        return self.__get_with_offset(url, "runs", **kwargs)

    def get_run(self, run_id):
        """
        """
        url = f'get_run/{run_id}'
        response = self.client.send_get(url)
        logger.info(response)
        return response

    def get_runs_by_name(self, project_id, run_name):
        runs = self.get_runs(project_id)
        if not isinstance(runs, list): runs = []
        run = list(filter(lambda x: x.get('name')== run_name, runs))
        if run and len(run) > 1: 
            msg = f'Found more than 1 test runs with name {run_name} for project {project_id}'
            self.__handle_duplicates(msg)
        return run
    
    def add_run(self, project_id, suite_id, name, description, include_all=True, refs=None, case_ids=None):
        if case_ids and not isinstance(case_ids, list):
            logger.error(f'{case_ids} should be a list')
        run = self.get_runs_by_name(project_id, name)
        if run:
            if len(run) > 1:
                return {}
            return run[0]
        description = description
        include_all = include_all
        url = f'add_run/{project_id}'
        data = {
            "suite_id": suite_id,
            "name": name,
            "description": description,
            "include_all": include_all
            }
        if refs:
            data['refs'] = refs
        if case_ids and not include_all:
            data['case_ids'] = case_ids
        logger.info(f'Adding test run {name} for project {project_id} and suite {suite_id}')
        response = self.client.send_post(url, data=data)
        logger.info(response)
        return response

    def update_run(self, run_id, **kwargs):
        url = f'update_run/{run_id}'
        response = self.client.send_post(url, data=kwargs)
        logger.info(response)
        return response

    def close_run(self, run_id):
        url = f'close_run/{run_id}'
        #logger.info(f'Adding attachment file {filepath} to run {run_id}')
        response = self.client.send_post(url, data={})
        logger.info(response)
        return response

    # def delete_run(self, run_id):
    #     url = f'delete_run/{run_id}'
    #     #logger.info(f'Adding attachment file {filepath} to run {run_id}')
    #     response = self.client.send_post(url, data={})
    #     return response

    # RESULTS
    def get_results(self, test_id, **kwargs):
        url = f'get_results/{test_id}'
        return self.__get_with_offset(url, "results", **kwargs) 

    def get_results_for_case(self, run_id, case_id, **kwargs):
        url = f'get_results_for_case/{run_id}/{case_id}'
        return self.__get_with_offset(url, "results", **kwargs)

    def get_results_for_run(self, run_id, **kwargs):
        url = f'get_results_for_run/{run_id}'
        return self.__get_with_offset(url, 'results', **kwargs)

    def add_result_for_case(self, run_id, case_id, status_id=2, comment=None, elapsed=''):
        case = self.get_case(case_id)
        if elapsed == '0s':
            elapsed = ''
        if not case:
            raise AssertionError(f'Case {case_id} does not exist')
        logger.info(f'Adding status {status_id} to run/test {run_id}/{case_id} ({case.get("name")})')
        url = f'add_result_for_case/{run_id}/{case_id}'
        data = {
            "status_id": status_id,
            "comment": comment,
            "elapsed": elapsed
        }
        response = self.client.send_post(url, data=data)
        logger.info(response)
        return response
    
    def add_result(self):
        # TODO
        pass

    # ATTACHMENTS
    def add_attachment_to_test_run(self, run_id, filepath):
        url = f'add_attachment_to_run/{run_id}'
        logger.info(f'Adding attachment file {filepath} to run {run_id}')
        response = self.client.send_post(url, data=filepath)
        logger.info(response)
        return response

    def get_attachments_for_run(self, run_id:int, **kwargs):
        logger.info(f'Getting attachments for run {run_id}')
        url = f'get_attachments_for_run/{run_id}'
        return self.__get_with_offset(url, "attachments", **kwargs)
        
    def delete_attachment(self, attachment:dict):
        attachment_id = attachment.get('id')
        attachment_name = attachment.get('name')
        attachment_created_timestamp = attachment.get('created_on')
        attachment_created_date = datetime.datetime.fromtimestamp(attachment_created_timestamp)
        logger.info(
            f'Deleting from testrail attachment {attachment_id}:{attachment_name} created on {attachment_created_date}')
        return self.delete_attachment_by_id(attachment_id)

    def delete_attachment_by_id(self, attachment_id):
        url = f'delete_attachment/{attachment_id}'
        response = self.client.send_post(url, data=None)
        logger.info(response)
        return response

    def get_statuses(self):
        """
        """
        url = f'get_statuses/'
        response = self.client.send_get(url)
        logger.info(response)
        return response

    def set_status_on_test_run(self, status_id, **kwargs):
        """
        """
        statuses = self.get_statuses()
        status_ids = list(map(lambda x: x.get("id"), statuses))
        if str(status_id).lower() in ('pass', 'passed'):
            status_id = TEST_STATUS_PASSED
        elif str(status_id).lower() in ('fail', 'failed'):
            status_id = TEST_STATUS_FAILED
        elif str(status_id).isdigit() and int(status_id) == TEST_STATUS_UNTESTED:
            logger.error(f'Setting status to Untested is not supported')
            return {}
        else:
            if not (str(status_id).isdigit() and int(status_id) in status_ids):
                raise Exception(f'{status_id} is not a valid status id.' 
                f'Should be any of {status_ids} or one of the following strings (case insensitive):'  
                'pass, passed, fail, failed') 
        kwargs['status_id'] = status_id
        self.set_test_case_result(**kwargs)
        suite_info = self.test_case_result.suite_id or self.test_case_result.suite_name
        test_case_info = self.test_case_result.test_id or self.test_case_result.test_name
        section_info = self.test_case_result.section_id or self.test_case_result.section_name
        run_info = self.test_case_result.run_id or self.test_case_result.run_name
        if not self.test_case_result.project_id or not suite_info or not test_case_info or not section_info or not run_info:
            if not self.test_case_result.project_id:
                logger.error('Testrail project id missing!')
            elif not run_info:
                logger.error(f'Test run info is missing, you must provide either a run name or a run id')
            elif not suite_info:
                logger.error(f'Suite info is missing, you must provide either a suite name or a suite id')
            elif not section_info:
                logger.error(f'Section info is missing, you must provide either a section name or a section id')
            elif not test_case_info:
                logger.error(f'Test case info is missing, you must provide either a test name or a test id')
            else:
                logger.error(f'Unexpected error for {self.test_case_result.project_id}')
            return {}
        if self.test_case_result.suite_name:
            suite = self.add_suite(project_id=self.test_case_result.project_id,
                                           suite_name=self.test_case_result.suite_name,
                                           description=self.test_case_result.description)
            if not suite:
                return {}
            self.test_case_result.suite_id = suite.get("id")
        if self.test_case_result.section_name:
            section = self.add_section(name=self.test_case_result.section_name, 
                                                   project_id=self.test_case_result.project_id,
                                                   suite_id=self.test_case_result.suite_id,
                                                   description=self.test_case_result.description
                                                   )
            if not section:
                return {}
            self.test_case_result.section_id = section.get("id")
            self.test_case_result.section_name = section.get("name")
        if self.test_case_result.test_name:
            test = self.add_case(section_name=self.test_case_result.section_name,
                                 section_id=self.test_case_result.section_id,
                                                  project_id=self.test_case_result.project_id,
                                                  suite_id=self.test_case_result.suite_id,
                                                  title=self.test_case_result.test_name)
            if not test:
                return {}
            self.test_case_result.test_id = test.get("id")

        if self.test_case_result.run_name:
            run = self.add_run(project_id=self.test_case_result.project_id,
                                                suite_id=self.test_case_result.suite_id,
                                                name=self.test_case_result.run_name,
                                                description=self.test_case_result.description,
                                                include_all=self.test_case_result.include_all)
            if not run:
                return {}
            self.test_case_result.run_id = run.get("id")
        response = self.add_result_for_case(run_id=self.test_case_result.run_id,
                                            case_id=self.test_case_result.test_id,
                                            elapsed=self.test_case_result.elapsed,
                                            status_id=self.test_case_result.status_id,
                                            comment=self.test_case_result.comment
        )
        logger.info(response)
        return response
    
    def upload_robot_to_testrail(self, outputdir, run_id, extension="html", type="log"):
        """
        Keyword used for uploading robot framework report files to Testrail. 

        """        
        files = os.listdir(outputdir)
        html_files = list(filter(lambda x: x.endswith(f'.{extension}') and x.startswith(type), files))
        for log_file in html_files:
            log_path = Path(outputdir) / log_file
            print(f'Adding file {log_path} to run {run_id}')
            response = self.add_attachment_to_test_run(run_id, log_path)
            if response:
                logger.info(f'Successfully uploaded file {log_path} to test run {run_id}: {response}')
            else:
                logger.error(f'Error: {response}')
            return response
              
    def delete_attachments_cleanup(self, run_id, max_to_keep=7):
        attachments = self.get_attachments_for_run(run_id)
        attachments_sorted = sorted(attachments, key=lambda x: x.get("created_on"))
        deleted = []
        if len(attachments_sorted) > max_to_keep:
            to_delete = attachments_sorted[0:len(attachments) - max_to_keep]
            
            for attachment in to_delete:
                #print(f'Deleting attachment {attachment["name"]} from run {run_id}')
                logger.info(f'Deleting attachment {attachment["name"]} from run {run_id}')
                try:
                    response = self.delete_attachment(attachment)
                    deleted.append(attachment)
                except Exception as e:
                    logger.error(e)  
        logger.info(deleted)      
        return deleted


class DuplicateException(BaseException):
    pass


import random
import string

STATUS_VALUES = ["passes", "blocked", "untested", "retest", "failed", "custom_6"]
GET_STATUSES = [{"id": i+1, "name": STATUS_VALUES[i]} for i in range(len(STATUS_VALUES))]

def get_sentence(count):
    text = ''
    for i in range(count):
        text = f'{text} {get_random_string(random.randint(4, 10))}'  
    return text

def get_random_string(count=8):
    return ''.join(random.choice(string.ascii_lowercase) for i in range(count))

class BaseResponse:

    def __init__(self, path, size=250, offset=0, limit=250) -> None:
        self.offset = offset
        self.limit = limit
        self.size = min(size, limit)
        self._links = {
            "next": f"/api/v2/{path}&limit={size}&offset={offset+size}",
            "prev": None if offset==0 else f"/api/v2/{path}&limit={limit}&offset={offset}"
        }
    
    @property
    def get(self):
        return vars(self)

class ProjectsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_projects", size=size, offset=offset, limit=limit)
        self.projects = [
                    {"id": i+offset+1, "name": f"Project {i+offset+1}", 
                    "announcement": f"This is a demo project ", 
                    "show_announcement": False, 
                    "is_completed": False, 
                    "completed_on": None, 
                    "suite_mode": 3, 
                    "url": f"https://example.testrail.net/index.php?/projects/overview/{i+offset+1}"
                    } 
                    for i in range(min(size, limit))
                    ]

class SuitesResponse(BaseResponse):
    def __init__(self, size=250, offset=0, limit=250):
        self.suites = [
            {"id": i+100, "name": f"Suite {i+100}"} for i in range(min(size, limit))
                    ]

class SectionsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_sections", size=size, offset=offset, limit=limit)
        self.sections = [
                    {"id": i+1000, "name": f"Section {i+1000}"} for i in range(min(size, limit))
                    ]

class RunsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_runs", size=size, offset=offset, limit=limit)
        self.runs = [
	            { "id": i+offset+1, "name": f"Test run {i+offset+1}", 'test_id': i+offset+1001, 
                'status_id': random.randint(1, 5), 'created_on': random.randint(16700000, 16800000), 'assignedto_id': None, 
                'comment': get_sentence(random.randint(3, 6)), 'version': None, 'elapsed': f'{random.randint(1, 120)}s', 'defects': None, 
                'created_by': random.randint(35, 40), 'custom_step_results': None, 
                'custom_environment': None, 'attachment_ids': []
                } for i in range(min(size, limit))]
                
class ResultsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_results", size=size, offset=offset, limit=limit)
        self.results = [{"id": i+5000, "test_id": i+1000, 
                        "status_id": random.randint(1, 5),
                        "comment": get_sentence(random.randint(3, 6))} for i in range(min(size, limit))
                        ]
        
class CasesResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_cases", size=size, offset=offset, limit=limit)
        self.cases = [{"id": i+offset+1, "title": f'Case {i+offset+1}', 
            "section_id": 1000} for i in range(min(size, limit))]

class TestsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_tests", size=size, offset=offset, limit=limit)
        self.tests = [{"id": i+offset+1, "title": f'Test Case {i+offset+1}', 
            "status_id": random.randint(1, 5)} for i in range(min(size, limit))]

class AttachmentsResponse(BaseResponse):

    def __init__(self, size=250, offset=0, limit=250) -> None:
        super().__init__("get_tests", size=size, offset=offset, limit=limit)
        self.attachments = [{ "id": i+offset+1, "name": f"Image_{i+offset+1}.jpg", 
                      "size": random.randint(500, 2500), "created_on": random.randint(1578900, 1600000),
                     "project_id": 17, "case_id": random.randint(5000, 5030), 
                     "user_id": random.randint(1, 5), 
                     "result_id": None} 
                     for i in range(min(size, limit))]


"""TestRail API client
"""

import base64
import json
import requests


class Apiclient:
    def __init__(self, base_url):
        self.user = ''
        self.password = ''
        if not base_url.endswith('/'):
            base_url = f'{base_url}/'
        self.__url = f'{base_url}index.php?/api/v2/'  

    def send_get(self, uri, filepath=None):
        """Issue a GET request.
        """
        return self.__send_request('GET', uri, filepath)

    def send_post(self, uri, data):
        """Issue a POST request.
        """
        return self.__send_request('POST', uri, data)

    def __send_request(self, method, uri, data):
        url = self.__url + uri

        auth = str(base64.b64encode(bytes(f'{self.user}:{self.password}', 'utf-8')), 'ascii').strip()
        headers = {'Authorization': 'Basic ' + auth}

        if method == 'POST':
            if uri[:14] == 'add_attachment':    # add_attachment API method
                files = {'attachment': (open(data, 'rb'))}
                response = requests.post(url, headers=headers, files=files)
                files['attachment'].close()
            else:
                headers['Content-Type'] = 'application/json'
                payload = bytes(json.dumps(data), 'utf-8')
                response = requests.post(url, headers=headers, data=payload)
        else:
            headers['Content-Type'] = 'application/json'
            response = requests.get(url, headers=headers)

        if response.status_code > 201:
            try:
                error = response.json()
            except:     # response.content not formatted as JSON
                error = str(response.content)
            raise Exception(f'TestRail API returned HTTP {response.status_code} ({error})')
        else:
            if uri[:15] == 'get_attachment/':   # Expecting file, not JSON
                try:
                    open(data, 'wb').write(response.content)
                    return (data)
                except:
                    return ("Error saving attachment.")
            else:
                try:
                    return response.json()
                except: # Nothing to return
                    return {}





import unittest
import random
from unittest.mock import patch
from unittest.mock import Mock
from TestrailLibrary.TestrailLibrary import TestrailLibrary as TR, DuplicateException
from .mock_responses import *


apiclient = "testrail.APIClient"


class BasicTests(unittest.TestCase):     

    def setUp(self) -> None:
        self.tr = TR("dummy")
        return super().setUp()

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_project(self, mock_get):
        projects = ProjectsResponse(5)
        project = projects.projects[0]
        mock_get.return_value = project
        res = self.tr.get_project(project.get("id"))
        self.assertTrue(res)
        self.assertEqual(res.get("id"), project.get("id"))
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_projects(self, mock_get):
        projects = ProjectsResponse(5)
        mock_get.return_value = projects.get
        res = self.tr.get_projects()
        self.assertTrue(res)
        self.assertTrue(len(res)==5)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=[]))
    def test_get_project_when_empty_projects(self):
        res = self.tr.get_project(1)
        self.assertFalse(res)

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_project_by_name(self, mock_get):
        projects = ProjectsResponse(5)
        project = projects.projects[0]
        mock_get.return_value = projects.get
        project_name = project.get("name")
        res = self.tr.get_project_by_name(project_name)
        self.assertTrue(res)
        self.assertEqual(res.get('id'), project.get("id"))
        self.assertEqual(res.get('name'), project_name)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=ProjectsResponse(10).get))
    def test_get_project_by_name_when_name_no_exists(self):
        res = self.tr.get_project_by_name("some project")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=[]))
    def test_get_project_by_name_when_empty_projects(self):
        res = self.tr.get_project_by_name("some project")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={"message": "Error"}))
    def test_get_project_by_name_when_error_response(self):
        res = self.tr.get_project_by_name("some project")
        self.assertFalse(res)

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suite(self, mock_get):
        suites = SuitesResponse(10).suites
        mock_get.return_value = suites[0]
        res = self.tr.get_suite(suites[0].get("id"))
        self.assertTrue(res)
        self.assertEqual(res.get('id'), suites[0].get('id'))

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suite_by_name(self, mock_get):
        suites = SuitesResponse(2).suites
        mock_get.return_value = suites
        res = self.tr.get_suites_by_name(17, suites[0].get("name"))
        self.assertTrue(res)
        self.assertEqual(res[0].get('id'), suites[0].get('id'))
        self.assertEqual(res[0].get('name'), suites[0].get("name"))

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suite_by_name_when_no_exists(self, mock_get):
        suites = SuitesResponse(10).suites
        mock_get.return_value = suites
        res = self.tr.get_suites_by_name(17, "some suite")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=[]))
    def test_get_suite_by_name_when_empty(self):
        res = self.tr.get_suites_by_name(17, "some suite")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={"message":"error"}))
    def test_get_suite_by_name_when_error(self):
        res = self.tr.get_suites_by_name(17, "some suite")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suite_by_name_when_duplicates(self, mock_get):
        suites = SuitesResponse(10).suites
        suites.append(suites[0])
        mock_get.return_value = suites
        with self.assertRaises(DuplicateException): 
            res = self.tr.get_suites_by_name(17, suites[0].get("name"))
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suite_by_name_when_duplicates_and_expect(self, mock_get):
        suites = SuitesResponse(10).suites
        suites.append(suites[0])
        mock_get.return_value = suites
        self.tr.set_expect_duplicates(True)
        res = self.tr.get_suites_by_name(17, suites[0].get("name"))
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].get("name"), res[1].get("name"))
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_suites_in_project(self, mock_get):
        suites = SuitesResponse(10).suites
        mock_get.return_value = suites
        res = self.tr.get_suites_in_project(17)
        self.assertTrue(res)
        self.assertEqual(res, suites)

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_run(self, mock_get):
        runs = RunsResponse(10)
        mock_get.return_value = runs.runs[0]
        res = self.tr.get_run(1)
        self.assertTrue(res)
        self.assertEqual(res.get('id'), 1) 

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_run_by_name(self, mock_get):
        runs = RunsResponse(10)
        mock_get.return_value = runs.get
        run = runs.runs[random.randint(0, len(runs.runs)-1)]
        run_name = run.get('name') 
        res = self.tr.get_runs_by_name(17, run_name)
        self.assertTrue(res)
        self.assertEqual(res[0].get('name'), run_name) 
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(10).get))
    def test_get_run_by_name_when_not_exists(self):
        res = self.tr.get_runs_by_name(17, "test")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(0).get))
    def test_get_run_by_name_when_empty(self):
        res = self.tr.get_runs_by_name(17, "test")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={"message":"error"}))
    def test_get_run_by_name_when_error(self):
        res = self.tr.get_runs_by_name(17, "test")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_run_by_name_when_duplicates(self, mock_get):
        runs = RunsResponse(10)
        runs.runs.append(runs.runs[0])
        mock_get.return_value = runs.get
        with self.assertRaises(DuplicateException): 
            res = self.tr.get_runs_by_name(17, runs.runs[0].get("name"))
        
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(side_effect=[RunsResponse().get, 
                                                 RunsResponse(10, 250).get])) 
    def test_get_runs(self):
        res = self.tr.get_runs(17)
        self.assertTrue(res)
        self.assertTrue(len(res)==260)
        
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(50).get))
    def test_get_runs_with_size_and_no_limit(self):
        res = self.tr.get_runs(17)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(10, limit=50).get)) 
    def test_get_runs_with_size_smaller_than_limit(self):
        res = self.tr.get_runs(17, limit=50)
        self.assertTrue(res)
        self.assertTrue(len(res)==10)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(50, limit=50).get)) 
    def test_get_runs_with_size_equal_to_limit(self):
        res = self.tr.get_runs(17, limit=50)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(60, limit=50).get)) 
    def test_get_runs_with_size_greater_than_limit(self):
        res = self.tr.get_runs(17, limit=50)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(1).get)) 
    def test_get_runs_with_size_one(self):
        res = self.tr.get_runs(17)
        self.assertTrue(res)
        self.assertTrue(len(res)==1)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(1, limit=1).get)) 
    def test_get_runs_with_size_one_and_limit_one(self):
        res = self.tr.get_runs(17, limit=1)
        self.assertTrue(res)
        self.assertTrue(len(res)==1)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(50, offset=5).get)) 
    def test_get_runs_with_offset(self):
        res = self.tr.get_runs(17, offset=5)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)
        self.assertEqual(res[0].get('id'), 6)
        self.assertEqual(res[4].get('id'), 10)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(50, offset=5, limit=2).get)) 
    def test_get_runs_with_offset_and_limit(self):
        res = self.tr.get_runs(17, offset=5, limit=2)
        self.assertTrue(res)
        self.assertTrue(len(res)==2)
        self.assertEqual(res[0].get('id'), 6)
        self.assertEqual(res[1].get('id'), 7)
    
    

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_case_by_name(self, mock_get):
        cases_1 = CasesResponse()
        cases_2 = CasesResponse(10, 250)
        mock_get.side_effect=[cases_1.get, cases_2.get]
        res = self.tr.get_cases_by_name(17, 1000, 1000, cases_1.cases[0].get("title"))
        self.assertTrue(res)
        self.assertEqual(res[0].get('title'), cases_1.cases[0].get("title")) 
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(20).get))
    def test_get_case_by_name_when_no_exists(self):
        res = self.tr.get_cases_by_name(17, 1000, 1000, "Case 500")
        self.assertFalse(res)  
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(0).get))
    def test_get_case_by_name_when_empty(self):
        res = self.tr.get_cases_by_name(17, 1000, 1000, "Case 500")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={"message":"error"}))
    def test_get_case_by_name_when_error(self):
        res = self.tr.get_cases_by_name(17, 1000, 1000, "Case 500")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_case_by_name_when_duplicates(self, mock_get):
        cases = CasesResponse(10)
        new_cases = CasesResponse(1)
        cases.cases.append(new_cases.cases[0])
        mock_get.return_value = cases.get
        with self.assertRaises(DuplicateException): 
            res = self.tr.get_cases_by_name(17, 1000, 1000, cases.cases[0].get("title"))

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_case_by_name_when_duplicates_and_expect(self, mock_get):
        cases = CasesResponse(10)
        new_cases = CasesResponse(1)
        cases.cases.append(new_cases.cases[0])
        mock_get.return_value = cases.get
        self.tr.set_expect_duplicates(True)
        res = self.tr.get_cases_by_name(17, 1000, 1000, cases.cases[0].get("title"))
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].get("name"), res[1].get("name"))
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(side_effect=[CasesResponse().get, 
                                                CasesResponse(10, offset=250).get]))
    def test_get_cases(self):
        res = self.tr.get_cases(17, 1000)
        self.assertTrue(res)
        self.assertTrue(len(res)==260)
        
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(20).get))
    def test_get_cases_with_size_and_no_limit(self): 
        res = self.tr.get_cases(17, 1000)
        self.assertTrue(res)
        self.assertTrue(len(res)==20)
        self.assertEqual(res[0].get("id"), 1)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(20, limit=5).get))
    def test_get_cases_with_size_and_limit(self): 
        res = self.tr.get_cases(17, 1000, limit=5)
        self.assertTrue(res)
        self.assertTrue(len(res)==5)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(5, limit=10).get))
    def test_get_cases_with_size_smaller_than_limit(self): 
        res = self.tr.get_cases(17, 1000, limit=10)
        self.assertTrue(res)
        self.assertTrue(len(res)==5)
        self.assertEqual(res[0].get("id"), 1)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(5, limit=5).get))
    def test_get_cases_with_size_equal_to_limit(self): 
        res = self.tr.get_cases(17, 1000, limit=5)
        self.assertTrue(res)
        self.assertTrue(len(res)==5)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(20, limit=10).get))
    def test_get_cases_with_size_greater_than_limit(self): 
        res = self.tr.get_cases(17, 1000, limit=10)
        self.assertTrue(res)
        self.assertTrue(len(res)==10)

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(1).get))
    def test_get_cases_with_size_one(self): 
        res = self.tr.get_cases(17, 1000)
        self.assertTrue(res)
        self.assertTrue(len(res)==1)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(1, limit=1).get))
    def test_get_cases_with_size_one_and_limit_one(self): 
        res = self.tr.get_cases(17, 1000, limit=1)
        self.assertTrue(res)
        self.assertTrue(len(res)==1)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(50, offset=5).get)) 
    def test_get_cases_with_offset(self):
        res = self.tr.get_cases(17, 1000, offset=5)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)
        self.assertEqual(res[0].get('id'), 6)
        self.assertEqual(res[4].get('id'), 10)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=CasesResponse(50, offset=5, limit=2).get)) 
    def test_get_cases_with_offset_and_limit(self):
        res = self.tr.get_cases(17, 1000, offset=5, limit=2)
        self.assertTrue(res)
        self.assertTrue(len(res)==2)
        self.assertEqual(res[0].get('id'), 6)
        self.assertEqual(res[1].get('id'), 7)


    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(50).get))
    @patch(f'TestrailLibrary.{apiclient}.send_post', Mock(return_value=RunsResponse(1, offset=1000).runs[0]))
    def test_add_run(self):
        res = self.tr.add_run(17, 1000, name="Test run 1001", description="")
        self.assertTrue(res)
        self.assertEqual(res.get("name"), "Test run 1001")

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=RunsResponse(10).get))
    def test_add_run_when_exists(self):
        res = self.tr.add_run(17, 1000, name="Test run 1", description="")
        self.assertTrue(res)
        self.assertEqual(res.get('name'), "Test run 1") 

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_add_run_when_duplicates(self, mock_get):
        runs = RunsResponse(10)
        runs.runs.append(runs.runs[0])
        mock_get.return_value = runs.get
        with self.assertRaises(DuplicateException): 
            res = self.tr.add_run(17, 1000, name=runs.runs[0].get("name"), description="")
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_add_run_when_duplicates_and_expected(self, mock_get):
        runs = RunsResponse(10)
        runs.runs.append(runs.runs[0])
        mock_get.return_value = runs.get
        self.tr.set_expect_duplicates(True)
        res = self.tr.add_run(17, 1000, name=runs.runs[0].get("name"), description="")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_add_case(self, mock_get, mock_post):
        suites = SuitesResponse(2)
        cases = CasesResponse(10)
        sections = SectionsResponse(2)
        section_name = sections.sections[0].get("name")
        mock_post.return_value = suites.suites[0]
        mock_get.side_effect = [
            suites.suites[0], sections.get, cases.get
        ]
        suite_id = suites.suites[0].get('id')
        res = self.tr.add_case(17, suite_id, "New Case", section_name=section_name)
        self.assertTrue(res)
        self.assertEqual(res.get('id'), suites.suites[0].get('id')) 
        self.assertEqual(res.get('title'), suites.suites[0].get('title')) 
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')  
    def test_add_case_when_exists(self, mock_get):
        cases = CasesResponse(10)
        suites = SuitesResponse(1)
        title = cases.cases[0].get("title")
        sections = SectionsResponse(2)
        section_name = sections.sections[0].get("name")
        case_id = cases.cases[0].get("id")
        mock_get.side_effect = [
            suites.suites[0], sections.get, cases.get
        ]
        res = self.tr.add_case(17, suites.suites[0].get("id"), title, section_name=section_name)
        self.assertTrue(res)
        self.assertEqual(res.get("id"), case_id)

    @patch(f'TestrailLibrary.{apiclient}.send_get')  
    def test_add_case_when_duplicates(self, mock_get):
        cases = CasesResponse(10)
        suites = SuitesResponse(1)
        new_cases = CasesResponse(1)
        new_case = new_cases.cases[0]
        new_case["id"] = 11
        cases.cases.append(new_case)
        title = cases.cases[0].get("title")
        sections = SectionsResponse(2)
        section_name = sections.sections[0].get("name")
        mock_get.side_effect = [
            suites.suites[0], sections.get, cases.get
        ]
        with self.assertRaises(DuplicateException):
            res = self.tr.add_case(17, suites.suites[0].get("id"), title, section_name=section_name)
    

    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=ResultsResponse(50).get)) 
    def test_get_results_for_case(self):
        res = self.tr.get_results_for_case(1, 1)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=ResultsResponse(0).get)) 
    def test_get_results_for_case_when_zero(self):
        res = self.tr.get_results_for_case(1, 1)
        self.assertFalse(res)
        self.assertTrue(len(res)==0)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={})) 
    def test_get_results_for_case_when_empty(self):
        res = self.tr.get_results_for_case(1, 1)
        self.assertFalse(res)
        self.assertTrue(len(res)==0)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=ResultsResponse(50).get)) 
    def test_get_results_for_run(self):
        res = self.tr.get_results_for_run(1)
        self.assertTrue(res)
        self.assertTrue(len(res)==50)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value=ResultsResponse(0).get)) 
    def test_get_results_for_run_when_empty(self):
        res = self.tr.get_results_for_run(1)
        self.assertFalse(res)
        self.assertTrue(len(res)==0)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get', Mock(return_value={"message": "error"}))
    def test_get_results_for_run_when_error(self):
        res = self.tr.get_results_for_run(1)
        self.assertFalse(res)
        self.assertTrue(len(res)==0)

    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_add_result_to_test(self, mock_get, mock_post):
        cases = CasesResponse(10)
        results = ResultsResponse(1)
        result = results.results[0]
        mock_get.return_value = cases.cases[0]
        mock_post.return_value = result
        res = self.tr.add_result_for_case(1000, 5000, 1, comment=result.get("comment"))
        self.assertTrue(res)
        self.assertEqual(res, result)
       
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run(self, mock_get, mock_post):
        results = ResultsResponse(1)
        result = results.results[0]
        mock_post.return_value = result
        mock_get.side_effect = [GET_STATUSES, CasesResponse(10).get]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_id=1, section_id=1000,
            test_id=5000, run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_suite_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(1)
        result = results.results[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, suites.suites, CasesResponse(10).get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_name=suites.suites[0].get("name"), 
            section_id=1000, test_id=5000, run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_suite_and_section_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        sections = SectionsResponse(10)
        section = sections.sections[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, suites.suites, sections.get, CasesResponse(10).get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_name=suites.suites[0].get("name"), 
            section_name=section.get("name"), test_id=5000, run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_suite_and_section_and_test_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        sections = SectionsResponse(10)
        section = sections.sections[0]
        cases = CasesResponse(10)
        case = cases.cases[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, suites.suites, sections.get, cases.get, 
                                 sections.get, cases.get, cases.get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_name=suites.suites[0].get("name"), 
            section_name=section.get("name"), test_name=case.get("title"), run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_all_names(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        sections = SectionsResponse(10)
        section = sections.sections[0]
        cases = CasesResponse(10)
        case = cases.cases[0]
        runs = RunsResponse(5)
        run = runs.runs[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, suites.suites, sections.get, cases.get, 
                                 sections.get, cases.get, runs.get, cases.get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_name=suites.suites[0].get("name"), 
            section_name=section.get("name"), test_name=case.get("title"), run_name=run.get("name") )
        self.assertTrue(res)
        self.assertEqual(res, result)

    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_section_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        sections = SectionsResponse(10)
        section = sections.sections[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, sections.get, CasesResponse(10).get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_id=suites.suites[0].get("id"), 
            section_name=section.get("name"), test_id=5000, run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_test_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        cases = CasesResponse(10)
        case = cases.cases[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, cases.get, SectionsResponse(10).get, cases.get, cases.get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_id=suites.suites[0].get("id"), 
            section_id=1000, test_name=case.get("title"), run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_run_name(self, mock_get, mock_post):
        results = ResultsResponse(1)
        suites = SuitesResponse(2)
        result = results.results[0]
        cases = CasesResponse(10)
        runs = RunsResponse(5)
        run = runs.runs[0]
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, runs.get, cases.get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_id=suites.suites[0].get("id"), 
            section_id=1000, test_id=5000, run_name=run.get("name") )
        self.assertTrue(res)
        self.assertEqual(res, result)

    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_test_name_when_duplicate(self, mock_get, mock_post):
        results = ResultsResponse(1)
        result = results.results[0]
        cases = CasesResponse(10)
        case = cases.cases[0]
        new_cases = CasesResponse(1, 10)
        new_case = new_cases.cases[0]
        new_case["title"] = case.get("title")
        cases.cases.append(new_case)
        mock_post.return_value = result
        mock_get.side_effect = [ GET_STATUSES, SectionsResponse(10).get, cases.get, cases.get ]
        res = self.tr.set_status_on_test_run(1, project_id=17, suite_id=100, 
            section_id=1000, test_name=case.get("title"), run_id=1 )
        self.assertTrue(res)
        self.assertEqual(res, result)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get') 
    def test_set_status_on_test_run_using_wrong_status_id(self, mock_get): 
        mock_get.return_value = GET_STATUSES
        with self.assertRaises(Exception):
            res = self.tr.set_status_on_test_run(8, project_id=17, suite_id=100, 
                section_id=1000, test_id=1, run_id=1 )
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    def test_update_run(self, mock_post):
        run = RunsResponse(1).runs[0]
        mock_post.return_value = run
        res = self.tr.update_run(1, description="New description")
        self.assertEqual(res, run)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_test(self, mock_get):
        test = CasesResponse(1).cases[0]
        mock_get.return_value = test
        res = self.tr.get_test(1)
        self.assertEqual(res, test)

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_tests(self, mock_get):
        mock_get.return_value = TestsResponse(10).get
        res = self.tr.get_tests(1)
        self.assertEqual(len(res), 10)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_tests_by_case_name(self, mock_get):
        tests = TestsResponse(10)
        test = tests.tests[0]
        mock_get.return_value = tests.get
        res = self.tr.get_tests_by_case_name(1, test.get("title"))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], test)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_tests_by_case_name_when_not_found(self, mock_get):
        tests = TestsResponse(10)
        test = tests.tests[0]
        mock_get.return_value = tests.get
        res = self.tr.get_tests_by_case_name(1, "New Test")
        self.assertFalse(res)
    
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    def test_add_attachment_to_test_run(self, mock_post):
        run = RunsResponse(1).runs[0]
        mock_post.return_value = {"attachment_id": 12345}
        res = self.tr.add_attachment_to_test_run(run_id=run.get("id"), filepath="filepath")
        self.assertTrue(res)
        self.assertEqual(res.get("attachment_id"), 12345)

    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_attachments_for_run(self, mock_get):
        run = RunsResponse(1).runs[0]
        mock_get.return_value = AttachmentsResponse(10).get
        res = self.tr.get_attachments_for_run(run_id=run.get("id"))
        self.assertEqual(len(res), 10)
    
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_get_attachments_for_run_when_not_found(self, mock_get):
        run = RunsResponse(1).runs[0]
        mock_get.return_value = AttachmentsResponse(0).get
        res = self.tr.get_attachments_for_run(run_id=run.get("id"))
        self.assertEqual(len(res), 0)
    
    @patch('os.listdir')
    @patch(f'TestrailLibrary.{apiclient}.send_post')
    def test_upload_robot_to_testrail(self, mock_post, mock_listdir):
        info = {"run_id": 1, "project_id": 17}
        #text = io.BytesIO(str(info).encode('utf8'))
        files = ["log.html", "report.html", "output.xml"]
        mock_listdir.return_value = files 
        run = RunsResponse(1).runs[0]
        mock_post.return_value = {"attachment_id": 12345}
        res = self.tr.upload_robot_to_testrail("", run.get("id"))
        self.assertEqual(res.get("attachment_id"), 12345)


    @patch(f'TestrailLibrary.{apiclient}.send_post')
    def test_delete_attachment(self, mock_post):
        attachment = AttachmentsResponse(1).attachments[0]
        mock_post.return_value = {"attachment_id": attachment.get("id")}
        res = self.tr.delete_attachment(attachment)
        self.assertEqual(res.get("attachment_id"), attachment.get("id"))

    @patch(f'TestrailLibrary.{apiclient}.send_post')
    @patch(f'TestrailLibrary.{apiclient}.send_get')
    def test_delete_attachments_cleanup(self, mock_get, mock_post):
        run = RunsResponse(1).runs[0]
        attachments = AttachmentsResponse(10)
        mock_get.return_value = attachments.get 
        attachments_sorted = sorted(attachments.attachments, key=lambda x: x.get("created_on"))
        expected_response = [attachments_sorted[i] for i in range(len(attachments_sorted[:-7]))]
        deleted_response = None
        mock_post.return_value = deleted_response
        res = self.tr.delete_attachments_cleanup(run.get("id"))
        self.assertEqual(res, expected_response)


if __name__ == "__main__":
    unittest.main()
        

